;von Noobody, fr projektspezifische Zwecke von "Rock me Amadeus" etwas modifiziert

Type TEntity
	Field CenterX#
	Field CenterY#
	
	Field Mass#
	Field Inertia#
	
	Field VX#
	Field VY#
	
	Field AX#
	Field AY#
	
	Field Omega#
	Field AngularV#
	
	Field VertexCount
	Field Vertex.TVertex[64]
	
	Field ForceCount
	Field Force.TForce[64]
	
	Field Locked
	Field Collided
	Field Deactivate
End Type

Type TVertex
	Field Mass#
	Field Distance#
	Field Angle#
	
	Field GlobalX#
	Field GlobalY#
	
	Field Parent.TEntity
End Type

Type TForce
	Field X#
	Field Y#
	
	Field FX#
	Field FY#
End Type

Global ProjectedMin#,ProjectedMax#

Function RadToDeg#(Radian#)
	Return Radian#*180/Pi
End Function

Function CreateVertex(Entity.TEntity,X#,Y#,Mass#)
	Local Vertex.TVertex=New TVertex
	Vertex\GlobalX#=X#
	Vertex\GlobalY#=Y#
	Vertex\Mass#=Mass#
	Vertex\Parent=Entity
	
	Entity\Vertex[Entity\VertexCount]=Vertex
	Entity\VertexCount=Entity\VertexCount+1
End Function

Function InitEntity(Entity.TEntity)
	For i=0 To Entity\VertexCount-1
		Entity\Mass#=Entity\Mass#+Entity\Vertex[i]\Mass#
		Entity\CenterX#=Entity\CenterX#+Entity\Vertex[i]\GlobalX#*Entity\Vertex[i]\Mass#
		Entity\CenterY#=Entity\CenterY#+Entity\Vertex[i]\GlobalY#*Entity\Vertex[i]\Mass#
	Next
	
	Entity\CenterX#=Entity\CenterX#/Entity\Mass#
	Entity\CenterY#=Entity\CenterY#/Entity\Mass#
	
	For i=0 To Entity\VertexCount-1
		Local DistX#=(Entity\Vertex[i]\GlobalX#-Entity\CenterX#)
		Local DistY#=(Entity\Vertex[i]\GlobalY#-Entity\CenterY#)
		
		Local DistSQ#=DistX#*DistX#+DistY#*DistY#
		
		Entity\Inertia#=Entity\Inertia#+Entity\Vertex[i]\Mass#*DistSQ#
		
		Entity\Vertex[i]\Distance#=Sqr(DistSQ#)
		Entity\Vertex[i]\Angle#=360-ATan2(DistY#,DistX#)
		
	Next
End Function

Function DotP#(X1#,Y1#,X2#,Y2#)
	Return X1#*X2#+Y1#*Y2#
End Function

Function UpdateEntities(Timestep#=1)
	Local Entity.TEntity
	For Entity.TEntity=Each TEntity
		If Not Entity\Deactivate Then
			If Not Entity\Locked Then
				Entity\Collided=False
				
				Local Torque#=0,ForceX#=0,ForceY#=0
				
				For i=0 To Entity\VertexCount
					Local VertIndex=i Mod Entity\VertexCount
					
					Local X#=Entity\CenterX#+Entity\Vertex[VertIndex]\Distance#*Cos(Entity\Omega#+Entity\Vertex[VertIndex]\Angle#)
					Local Y#=Entity\CenterY#-Entity\Vertex[VertIndex]\Distance#*Sin(Entity\Omega#+Entity\Vertex[VertIndex]\Angle#)
					
					Entity\Vertex[VertIndex]\GlobalX#=X#
					Entity\Vertex[VertIndex]\GlobalY#=Y#
					
					;DEBUG - Fr Auto-Koll
					If KeyDown(57) Then
						Color 255,0,0
						Oval X-2-scroll_x,Y-2-scroll_y,4,4,1
					EndIf
				Next
				
				For i=0 To Entity\ForceCount-1
					Local VX#=Entity\CenterX#-Entity\Force[i]\X#
					Local VY#=Entity\CenterY#-Entity\Force[i]\Y#
					
					Local Tau#=VX#*Entity\Force[i]\FY#-VY#*Entity\Force[i]\FX#
					
					Torque#=Torque#+Tau#/Entity\Inertia#
					ForceX#=ForceX#+Entity\Force[i]\FX#
					ForceY#=ForceY#+Entity\Force[i]\FY#
					
					Delete Entity\Force[i]
				Next
				
				Entity\VX#=Entity\VX#+Timestep#*ForceX#/Entity\Mass#
				Entity\VY#=Entity\VY#+Timestep#*ForceY#/Entity\Mass#
				
				Entity\CenterX#=Entity\CenterX#+Entity\VX#*Timestep#
				Entity\CenterY#=Entity\CenterY#+Entity\VY#*Timestep#
				
				Entity\AngularV#=Entity\AngularV#+Torque#*Timestep#
				Entity\Omega#=Entity\Omega#+RadToDeg(Entity\AngularV#)*Timestep#
				
				Entity\ForceCount=0
			EndIf
		EndIf
	Next
	
	;Ursprngliche Koll-Prfung
	;Wenn diese Methode verwendet wird, sollte zumindest die tile-Abfrage in "processCollision" aktiviert werden!
;	Local E1.TEntity
;	Local E2.TEntity
;	For E1=Each TEntity
;		If Not E1\Deactivate Then
;			E2=After E1
;			While E2<>Null
;				If ((Not E1\Locked) Or (Not E2\Locked)) And (Not E2\Deactivate) Then
;					ProcessCollision E1,E2
;					count=count+1
;				EndIf
;				E2=After E2
;			Wend
;		EndIf
;	Next
	
	;Laufzeitoptimiert (speziell fr "Rock me Amadeus" geschrieben - nicht universiell anwendbar)
	;CPU-Ersparnis ist auf greren Maps hher als auf kleinen
	;Prinzip: Nur Tiles abfragen, die in der Umgebung der Player sind,
	;         Kollisions-Entities dorthin verschieben
	Local player.player,pl.player
	Local tile.tile
	Local xx,yy
	For player=Each player
		X=player\body\CenterX/tile_size
		Y=player\body\CenterY/tile_size
		For xx=X-2 To X+1
			If xx>=0 And xx<=map_w-1 Then
				For yy=Y-2 To Y+1
					If yy>=0 And yy<=map_h-1 Then
						tile=map(xx,yy)
						If tile<>Null Then
							For i=0 To 31
								If tile\PhysEntity[i]=Null Then Exit
								
								MovePhysEntity(tile\PhysEntity[i],xx*tile_size,yy*tile_size)
								
								ProcessCollision player\body,tile\PhysEntity[i]
								
								MovePhysEntity(tile\PhysEntity[i],-xx*tile_size,-yy*tile_size)
							Next
						EndIf
					EndIf
				Next
			EndIf
		Next
		pl=After player
		While pl<>Null
			ProcessCollision player\body,pl\body
			pl=After pl
		Wend
	Next
End Function

Function AddForce(Entity.TEntity,X#,Y#,FX#,FY#)
	Local Force.TForce=New TForce
	Force\X#=X#
	Force\Y#=Y#
	Force\FX#=FX#
	Force\FY#=FY#
	
	Entity\Force[Entity\ForceCount]=Force
	Entity\ForceCount=Entity\ForceCount+1
End Function

Function ProcessCollision(E1.TEntity,E2.TEntity)
	;Tile-Abfrage zur Performance-Steigerung (Nur bei Rock me Amadeus anwendbar!)
	;Local dist#=(E1\CenterX-E2\CenterX)*(E1\CenterX-E2\CenterX)+(E1\CenterY-E2\CenterY)*(E1\CenterY-E2\CenterY)
	;If dist>2*tile_size*tile_size Then Return False
	
	Local MinDistance#=10000000,CollV1.TVertex,CollV2.TVertex
	For i=0 To E1\VertexCount+E2\VertexCount-1
		If i < E1\VertexCount Then
			V1.TVertex=E1\Vertex[i]
			V2.TVertex=E1\Vertex[(i+1) Mod E1\VertexCount]
		Else
			V1.TVertex=E2\Vertex[i-E1\VertexCount]
			V2.TVertex=E2\Vertex[(i-E1\VertexCount+1) Mod E2\VertexCount]
		EndIf
		
		Local AxisX#=V1\GlobalY#-V2\GlobalY#
		Local AxisY#=V2\GlobalX#-V1\GlobalX#
		TFormNormal AxisX#,AxisY#,0,0,0
		AxisX#=TFormedX()
		AxisY#=TFormedY()
		
		ProjectToAxis(AxisX#,AxisY#,E1)
		Local Min#=ProjectedMin#
		Local Max#=ProjectedMax#
		ProjectToAxis(AxisX#,AxisY#,E2)
		
		Local Distance#=IntervalDistance(Min#,Max#,ProjectedMin#,ProjectedMax#)
		If Distance# > 0 Then
			Return False
		ElseIf Abs(Distance#) < MinDistance# Then
			MinDistance#=Abs(Distance#)
			
			Local CollAxisX#=AxisX#
			Local CollAxisY#=AxisY#
			
			Local DiffX#=E1\CenterX#-E2\CenterX#
			Local DiffY#=E1\CenterY#-E2\CenterY#
			
			If DotP(CollAxisX#,CollAxisY#,DiffX#,DiffY#) < 0 Then
				DiffX#=-DiffX#
				DiffY#=-DiffY#
			EndIf
			
			CollV1=V1
			CollV2=V2
		EndIf
	Next
	
	Local CollisionVX#=CollAxisX#*MinDistance#
	Local CollisionVY#=CollAxisY#*MinDistance#
	
	If CollV1\Parent <> E2 Then
		Local Temp.TEntity=E2
		E2=E1
		E1=Temp
	EndIf
	
	Local C#=-E2\CenterX#*CollAxisX#-E2\CenterY#*CollAxisY#
	
	Local Sign=Sgn(E1\CenterX#*CollAxisX#+E1\CenterY#*CollAxisY#+C#)
	
	If Sign=1 Then
		CollisionVX#=-CollisionVX#
		CollisionVY#=-CollisionVY#
	Else
		CollAxisX#=-CollAxisX#
		CollAxisY#=-CollAxisY#
	EndIf
	
	Local SmallestY#=1000000
	For i=0 To E1\VertexCount-1
		Local Vertex.TVertex=E1\Vertex[i]
		
		Local D#=Vertex\GlobalX#*CollAxisX#+Vertex\GlobalY#*CollAxisY#+C#
		
		If D# < SmallestY# Then
			SmallestY#=D#
			Local LowestVertex.TVertex=Vertex
		EndIf
	Next
	
	CollisionPX#=LowestVertex\GlobalX#
	CollisionPY#=LowestVertex\GlobalY#
	
	If DotP(CollisionVX#,CollisionVY#,E1\CenterX#-E2\CenterX#,E1\CenterY#-E2\CenterY#) < 0 Then
		CollisionVX#=-CollisionVX#
		CollisionVY#=-CollisionVY#
	EndIf
	
	Local RX1#=CollisionPY#-E1\CenterY#
	Local RY1#=-(CollisionPX#-E1\CenterX#)
	
	Local VX1#=E1\VX#+RX1#*E1\AngularV#
	Local VY1#=E1\VY#+RY1#*E1\AngularV#
	
	Local RX2#=CollisionPY#-E2\CenterY#
	Local RY2#=-(CollisionPX#-E2\CenterX#)
	
	Local VX2#=E2\VX#+RX2#*E2\AngularV#
	Local VY2#=E2\VY#+RY2#*E2\AngularV#
	
	Local VX#=VX1#-VX2#
	Local VY#=VY1#-VY2#
	
	Local NX#=CollAxisX#
	Local NY#=CollAxisY#
	
	Local PS1#=RX1#*NX#+RY1#*NY#
	Local PS2#=RX2#*NX#+RY2#*NY#
	
	Local E#=1
	
	If VX#*NX#+VY#*NY# < 0 Then
		Local J#
		If E1\Locked Then
			J#=-(1+E)*(NX#*VX#+NY#*VY#)/(1.0/E2\Mass#+PS2#*PS2#/E2\Inertia#)
		ElseIf E2\Locked Then
			J#=-(1+E)*(NX#*VX#+NY#*VY#)/(1.0/E1\Mass#+PS1#*PS1#/E1\Inertia#)
		Else
			J#=-(1+E)*(NX#*VX#+NY#*VY#)/(1.0/E1\Mass#+1.0/E2\Mass#+PS1#*PS1#/E1\Inertia#+PS2#*PS2#/E2\Inertia#)
		EndIf
		
		If Not E1\Locked Then
			E1\Collided=True
			E1\VX#=E1\VX#+NX#*J#/E1\Mass#
			E1\VY#=E1\VY#+NY#*J#/E1\Mass#
			E1\AngularV#=E1\AngularV#+(RX1#*NX#*J#+RY1#*NY#*J#)/E1\Inertia#
		EndIf
		
		If Not E2\Locked Then
			E2\Collided=True
			E2\VX#=E2\VX#-NX#*J#/E2\Mass#
			E2\VY#=E2\VY#-NY#*J#/E2\Mass#
			E2\AngularV#=E2\AngularV#-(RX2#*NX#*J#+RY2#*NY#*J#)/E2\Inertia#
		EndIf
	EndIf
	
	Local Ratio1#,Ratio2#
	If E2\Locked Then
		Ratio1#=1.0
		Ratio2#=0.0
	ElseIf E1\Locked Then
		Ratio1#=0.0
		Ratio2#=1.0
	Else
		Ratio1#=E2\Mass#/(E2\Mass#+E1\Mass#)
		Ratio2#=E1\Mass#/(E2\Mass#+E1\Mass#)
	EndIf
	
	E1\CenterX#=E1\CenterX#+Ratio1#*CollisionVX#
	E1\CenterY#=E1\CenterY#+Ratio1#*CollisionVY#
	
	E2\CenterX#=E2\CenterX#-Ratio2#*CollisionVX#
	E2\CenterY#=E2\CenterY#-Ratio2#*CollisionVY#
	
	Return True
End Function

Function ProjectToAxis(AxisX#,AxisY#,Entity.TEntity)
	Local DotP#=DotP(AxisX#,AxisY#,Entity\Vertex[0]\GlobalX#,Entity\Vertex[0]\GlobalY#)
	ProjectedMin#=DotP#
	ProjectedMax#=DotP#
	
	For i=1 To Entity\VertexCount-1
		DotP#=DotP(AxisX#,AxisY#,Entity\Vertex[i]\GlobalX#,Entity\Vertex[i]\GlobalY#)
		
		If DotP# < ProjectedMin# Then ProjectedMin#=DotP# ElseIf DotP# > ProjectedMax# Then ProjectedMax#=DotP#
	Next
End Function

Function IntervalDistance#(A1#,B1#,A2#,B2#)
	If A1# < A2# Then
		Return A2#-B1#
	Else
		Return A1#-B2#
	EndIf
End Function

;Physik-Erweiterung
Function PositionPhysEntity(Entity.TEntity,x#,y#)
	Entity\CenterX#=x#
	Entity\CenterY#=y#
End Function

Function MovePhysEntity(E.TEntity,x#,y#)
	For i=0 To E\VertexCount-1
		E\Vertex[i]\GlobalX#=E\Vertex[i]\GlobalX+x
		E\Vertex[i]\GlobalY#=E\Vertex[i]\GlobalY+y
	Next
	E\CenterX=E\CenterX+x
	E\CenterY=E\CenterY+y
End Function

Function RemoveEntity(Entity.TEntity)
	For i=0 To Entity\VertexCount-1
		Delete Entity\Vertex[i]
	Next
	For i=0 To Entity\ForceCount-1
		Delete Entity\Force[i]
	Next
	Delete Entity
End Function

Function CopyPhysEntity.TEntity(Entity.TEntity)
	Local E1.TEntity=New TEntity
	
	For i=0 To Entity\VertexCount-1
		CreateVertex(E1,Entity\Vertex[i]\GlobalX,Entity\Vertex[i]\GlobalY,1/Float(Entity\VertexCount))
	Next
	
	InitEntity(E1)
	Return E1
End Function
;~IDEal Editor Parameters:
;~F#2#1D#28#32#36#41#59#5D#C8#D3#171#17D#186#18B#194#19E
;~C#Blitz3D